from qcanvas.QtVersionHelper import QT_VERSION

if QT_VERSION == 5:
    # noinspection PyUnresolvedReferences
    from PySide2.QtCore import *
elif QT_VERSION == 6:
    # noinspection PyUnresolvedReferences
    from PySide6.QtCore import *



import traceback
from threading import Semaphore
from typing import Optional

from PySide6.QtWidgets import QProgressBar
from qasync import asyncSlot

from qcanvas.QtVersionHelper.QtCore import Slot, QUrl
from qcanvas.QtVersionHelper.QtGui import QDesktopServices
from qcanvas.QtVersionHelper.QtWidgets import QDialog, QWidget, QHBoxLayout, QLabel, QLineEdit, QVBoxLayout, \
    QDialogButtonBox, QGridLayout, QMessageBox
from qcanvas.net.canvas import CanvasClient
from qcanvas.util import AppSettings

tutorial_url = "https://www.iorad.com/player/2053777/Canvas---How-to-generate-an-access-token-"

def row(name: str) -> QWidget:
    widget = QWidget()
    layout = QHBoxLayout()

    layout.addWidget(QLabel(name))
    layout.addWidget(QLineEdit())

    widget.setLayout(layout)

    return widget


class SetupDialog(QDialog):
    def __init__(self, parent: Optional[QWidget] = None, allow_cancel: bool = True):
        super().__init__(parent)
        self.setWindowTitle("Setup")
        self._row_counter = 0
        self._operation_sem = Semaphore()

        self.progress_bar = QProgressBar()
        self.progress_bar.setMaximum(0)
        self.progress_bar.setMinimum(0)
        self.progress_bar.setValue(0)
        self.progress_bar.hide()

        self.grid = QGridLayout()
        grid_widget = QWidget()
        grid_widget.setLayout(self.grid)

        stack = QVBoxLayout()

        self.canvas_url = QLineEdit(AppSettings.canvas_url or "")
        # self.panopto_url = QLineEdit()
        self.canvas_api_key = QLineEdit(AppSettings.canvas_api_key or "")

        self._row("Canvas URL", self.canvas_url)
        # self._row("Painopto URL", self.panopto_url)
        self._row("Canvas API key", self.canvas_api_key)

        self.grid.addWidget(self.progress_bar, self._row_counter, 0, 1, 2)

        button_box = QDialogButtonBox(
            QDialogButtonBox.Save | QDialogButtonBox.Cancel if allow_cancel else QDialogButtonBox.Ok)
        button_box.addButton("How to get a canvas API key?", QDialogButtonBox.ButtonRole.HelpRole)

        button_box.helpRequested.connect(self._show_help)
        button_box.accepted.connect(self._verify)
        button_box.rejected.connect(lambda: self.reject())

        stack.addWidget(grid_widget)
        stack.addWidget(button_box)

        self.setLayout(stack)
        self.resize(500, 200)

    @asyncSlot()
    async def _verify(self):
        if self._operation_sem.acquire(False):
            try:
                def invalid(name: str):
                    msg = QMessageBox(QMessageBox.Icon.Warning,
                                      "Error",
                                      f"{name} is invalid",
                                      parent=self
                    )
                    msg.show()

                def ensure_protocol(url: str):
                    if len(url) > 0 and not (url.startswith("http://") or url.startswith("https://")):
                        return "https://" + url
                    else:
                        return url

                canvas_url_text = ensure_protocol(self.canvas_url.text().strip())
                # panopto_url_text = ensure_protocol(self.panopto_url.text().strip())
                canvas_api_key_text = self.canvas_api_key.text().strip()

                if not (len(canvas_url_text) > 0 and QUrl(canvas_url_text).isValid()):
                    invalid("Canvas URL")
                    return
                # if not (len(panopto_url_text) > 0 and QUrl(panopto_url_text).isValid()):
                #     invalid("Panopto URL")
                #     return
                elif not len(canvas_api_key_text) > 0:
                    invalid("API key")
                elif not (await self._verify_canvas_config(canvas_url_text, canvas_api_key_text)):
                    msg = QMessageBox(QMessageBox.Icon.Warning,
                                      "Error",
                                      f"The canvas URL or API key is invalid.\nPlease check you entered them correctly.",
                                      parent=self
                    )
                    msg.show()
                else:
                    AppSettings.canvas_url = canvas_url_text
                    AppSettings.canvas_api_key = canvas_api_key_text

                    self.accept()
            finally:
                self._operation_sem.release()
        else:
            QMessageBox(QMessageBox.Icon.Critical, "Error", "An operation is already in progress", parent=self).show()

    async def _verify_canvas_config(self, canvas_url: str, api_key: str) -> bool:
        self.progress_bar.show()

        try:
            return await CanvasClient.verify_config(canvas_url, api_key)
        except:
            traceback.print_exc()
            return False
        finally:
            self.progress_bar.hide()

    @Slot()
    def _show_help(self):
        msg = QMessageBox(
            QMessageBox.Icon.Information,
            "Help",
            """An interactive tutorial will open in your browser when you click OK.
            
Note that the "purpose" text doesn't matter and you can enter anything you want.

You should also leave the "expires" item blank if you want the key to last forever.

Don't share this key. You can revoke it at any time.""",
            parent=self
        )
        msg.accepted.connect(lambda: QDesktopServices.openUrl(tutorial_url))
        msg.show()


    def _row(self, name: str, widget: QWidget):
        self.grid.addWidget(QLabel(name), self._row_counter, 0)
        self.grid.addWidget(widget, self._row_counter, 1)

        self._row_counter += 1



from typing import Any

from qcanvas.QtVersionHelper import QT_VERSION

if QT_VERSION == 5:
    # noinspection PyUnresolvedReferences
    from PySide2.QtGui import *

    # Seemed to have moved packages... make things easier for myself
    # noinspection PyUnresolvedReferences
    from PySide2.QtWidgets import QAction, QActionGroup
elif QT_VERSION == 6:
    # noinspection PyUnresolvedReferences
    from PySide6.QtGui import *


def create_qaction(name: str, shortcut: QKeySequence | None = None, parent: Any = None, triggered: Any = None,
                   checkable: bool | None = None, checked : bool | None = None) -> QAction:
    action = QAction(name)

    if shortcut is not None:
        action.setShortcut(shortcut)

    if parent is not None:
        action.setParent(parent)

    if triggered is not None:
        action.triggered.connect(triggered)

    if checkable is not None:
        action.setCheckable(checkable)

        if checked is not None:
            action.setChecked(checked)

    return action



import asyncio
from typing import Callable, Any

from qcanvas.QtVersionHelper.QtCore import QObject, Signal
from qcanvas.util.task_pool import TaskPool


class DownloadPool(TaskPool[None], QObject):
    """
    Like the TaskPool, but uses a queue (a channel in kotlin terms) in the task to report download progress.
    When using .submit(), it is EXPECTED that the lambda returns a function that uses yield to signal progress, e.g:
    ```
    pool.submit(my_task_id, lambda: my_func())

    async def my_func(channel: asyncio.Queue):
        for i in range(0, 10):
            await slow_thing()
            yield (i / 10) # can be whatever format/type you want
    ```
    """
    download_progress_updated = Signal(object, Any)
    download_failed = Signal(object)
    download_finished = Signal(object)

    def __init__(self, parent: QObject | None = None):
        QObject.__init__(self, parent)
        TaskPool.__init__(self, echo=True)

        self._download_progress: dict[object, Any] = {}
        self._download_progress_sem = asyncio.Semaphore()

    async def _handle_task(self, func: Callable, task_id: object, event: asyncio.Event, func_args: dict):
        sem = self._semaphore

        try:
            async for progress in func(**func_args):
                async with self._download_progress_sem:
                    self._download_progress[task_id] = progress

                self.download_progress_updated.emit(task_id, progress)
        except BaseException as e:
            # Try to maintain integrity when a task fails
            async with self._download_progress_sem:
                if task_id in self._download_progress:
                    del self._download_progress[task_id]

            async with sem:
                event.set()
                del self._results[task_id]

            self.download_failed.emit(task_id)
            raise e

        await sem.acquire()

        if self._echo: print(f"Task {task_id} finished.")

        self._results[task_id] = None
        event.set()
        sem.release()

        self.download_finished.emit(task_id)

        return None

    async def get_task_progress(self, task_id: object):
        async with self._download_progress_sem:
            if task_id in self._download_progress:
                return self._download_progress[task_id]
            else:
                return None



import sys

QT_VERSION = 6

# if sys.modules.get("PySide6.QtCore"):
#     QT_VERSION = 6
# elif sys.modules.get("PySide2.QtCore"):
#     QT_VERSION = 2
# else:
#     QT_VERSION = None
#     raise Exception("Unsupported version of qt")
#
# def set_qt_version():
#     QT_VERSION = 6

def run_app(app):
    if sys.modules.get("PySide6.QtCore"):
        return app.exec()
    else:
        return app.exec_()


import logging
from asyncio import Event
from typing import Sequence, Optional

from qasync import asyncSlot

import qcanvas.db.database as db
from qcanvas.QtVersionHelper.QtCore import Slot, Signal, Qt, QUrl
from qcanvas.QtVersionHelper.QtGui import QDesktopServices
from qcanvas.QtVersionHelper.QtWidgets import *
from qcanvas.ui.viewer.course_list import CourseList
from qcanvas.ui.viewer.file_list import FileRow
from qcanvas.ui.viewer.file_view_tab import FileViewTab
from qcanvas.ui.viewer.page_list_viewer import AssignmentsViewer, PagesViewer, LinkTransformer
from qcanvas.util import AppSettings
from qcanvas.util.course_indexer import DataManager

_aux_settings = AppSettings.auxiliary

class AppMainWindow(QMainWindow):
    logger = logging.getLogger()
    loaded = Signal()
    operation_lock = Event()

    def __init__(self, data_manager: DataManager, parent: QWidget | None = None):
        super().__init__(parent)

        self.selected_course: db.Course | None = None
        self.courses: Sequence[db.Course] = []
        self.resources: dict[str, db.Resource] = {}
        self.data_manager = data_manager
        self.link_transformer = LinkTransformer(self.data_manager.link_scanners, self.resources)

        right_splitter = QSplitter()
        right_splitter.setOrientation(Qt.Orientation.Vertical)

        self.sync_button = QPushButton("Synchronize")
        self.sync_button.clicked.connect(self.sync_data)

        self.course_list = CourseList(self.data_manager)
        self.course_list.course_selected.connect(self.on_course_selected)

        self.assignment_viewer = AssignmentsViewer(self.link_transformer)
        self.assignment_viewer.viewer.anchorClicked.connect(self.viewer_link_clicked)

        self.pages_viewer = PagesViewer(self.link_transformer)
        self.pages_viewer.viewer.anchorClicked.connect(self.viewer_link_clicked)

        self.file_viewer = FileViewTab(data_manager.download_pool)
        self.file_viewer.group_by_preference_changed.connect(self.course_file_group_by_preference_changed)

        self.file_viewer.files_column.tree.itemActivated.connect(self.download_file_from_file_pane)
        self.file_viewer.assignment_files_column.tree.itemActivated.connect(self.download_file_from_file_pane)

        self.tab_widget = QTabWidget()
        self.tab_widget.insertTab(0, self.file_viewer, "Files")
        self.tab_widget.insertTab(1, self.assignment_viewer, "Assignments")
        self.tab_widget.insertTab(2, self.pages_viewer, "Pages")

        h_layout = QHBoxLayout()
        h_layout.addWidget(self.course_list)
        h_layout.addWidget(self.tab_widget)
        h_layout.setStretch(1, 1)

        v_layout = QVBoxLayout()
        v_layout.addLayout(h_layout)
        v_layout.addWidget(self.sync_button)

        widget = QWidget()
        widget.setLayout(v_layout)
        self.setCentralWidget(widget)

        self.loaded.connect(self.load_course_list)
        self.loaded.emit()

        self.read_settings()

    def closeEvent(self, event):
        _aux_settings.setValue("geometry", self.saveGeometry())
        _aux_settings.setValue("windowState", self.saveState())

    def read_settings(self):
        self.restoreGeometry(_aux_settings.value("geometry"))
        self.restoreState(_aux_settings.value("windowState"))

    @asyncSlot(QUrl)
    async def viewer_link_clicked(self, url: QUrl):
        # The url of a transformed link will start with an '@'
        if url.toString().startswith(LinkTransformer.transformed_url_prefix):
            # The rest of the 'url' is just the file id
            resource = self.resources[url.toString().removeprefix(LinkTransformer.transformed_url_prefix)]

            await self.data_manager.download_resource(resource)
            QDesktopServices.openUrl(QUrl.fromLocalFile(resource.download_location.absolute()))


    @asyncSlot(QTreeWidgetItem, int)
    async def download_file_from_file_pane(self, item: QTreeWidgetItem, _ : int):
        if isinstance(item, FileRow):
            await self.data_manager.download_resource(item.resource)
            QDesktopServices.openUrl(QUrl.fromLocalFile(item.resource.download_location.absolute()))


    @asyncSlot()
    async def sync_data(self):
        # # self.operation_lock.
        self.sync_button.setEnabled(False)
        self.sync_button.setText("Synchronizing")
        try:
            await self.data_manager.synchronize_with_canvas()
            await self.load_course_list()

        finally:
            self.sync_button.setEnabled(True)
            self.sync_button.setText("Synchronize")



    @asyncSlot()
    async def load_course_list(self):
        self.courses = (await self.data_manager.get_data())
        self.selected_course = None
        self.resources.clear()

        for course in self.courses:
            self.resources.update({resource.id: resource for resource in course.resources})

        self.course_list.load_course_list(self.courses)


    @Slot(db.Course)
    def on_course_selected(self, course: Optional[db.Course]):
        if course is not None:
            self.selected_course = course
            self.pages_viewer.fill_tree(course)
            self.assignment_viewer.fill_tree(course)
            self.file_viewer.load_course_files(course)
        else:
            self.selected_course = None
            self.file_viewer.clear()


    @asyncSlot(db.CoursePreferences)
    async def course_file_group_by_preference_changed(self, preference: db.GroupByPreference):
        self.selected_course.preferences.files_group_by_preference = preference
        await self.data_manager.update_item(self.selected_course.preferences)
        self.file_viewer.load_course_files(self.selected_course)




from typing import Sequence

import qcanvas.db as db
from qcanvas.QtVersionHelper.QtCore import Signal, Qt, Slot, QPoint
from qcanvas.QtVersionHelper.QtGui import QActionGroup, create_qaction
from qcanvas.QtVersionHelper.QtWidgets import *  # QWidget, QTreeView, QGroupBox, QBoxLayout, QHeaderView, QHBoxLayout, QComboBox, QLabel
from qcanvas.ui.viewer.file_list import FileList
from qcanvas.util.constants import default_assignments_module_names
from qcanvas.util.download_pool import DownloadPool


class FileColumn(QGroupBox):
    def __init__(self, column_name, download_pool: DownloadPool):
        super().__init__(title=column_name)

        self.tree = FileList(download_pool)
        self.setLayout(QBoxLayout(QBoxLayout.Direction.TopToBottom))
        self.layout().addWidget(self.tree)

    def load_items(self, items: Sequence[db.ModuleItem | db.Module]):
        self.tree.load_items(items)

    def clear(self):
        self.tree.clear()


class FileViewTab(QWidget):
    group_by_preference_changed = Signal(db.GroupByPreference)

    def __init__(self, download_pool: DownloadPool):
        super().__init__()

        self.files_column = FileColumn("Files", download_pool)
        self.assignment_files_column = FileColumn("Assignment files", download_pool)

        self.group_by_combobox = QComboBox()

        self.group_by_preference: db.GroupByPreference | None = None
        self.group_preference_layout = QHBoxLayout()
        self.group_preference_layout.addWidget(QLabel("Group By:"))
        self.group_preference_layout.addWidget(self.group_by_combobox)
        self.group_preference_layout.setStretch(1, 1)

        widget = QWidget()
        widget.setLayout(self.group_preference_layout)

        # fixme this can't stay here
        self.files_column.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.files_column.customContextMenuRequested.connect(self.files_column_context_menu)

        self.group_by_preference_changed.connect(self.update_group_by_preferences)

        layout = QHBoxLayout()
        layout.addWidget(self.files_column)
        layout.addWidget(self.assignment_files_column)

        self.setLayout(layout)

    @Slot(db.GroupByPreference)
    def update_group_by_preferences(self, preference : db.GroupByPreference):
        if self.group_by_preference is not None:
            self.group_by_preference = preference

    def load_course_files(self, course: db.Course):
        self.group_by_preference = course.preferences.files_group_by_preference

        module_items: list[db.ModuleItem] = []
        assignment_items: list[db.ModuleItem] = []

        for module_item in course.module_items:
            if module_item.module.name.lower() in default_assignments_module_names:
                assignment_items.append(module_item)
            else:
                module_items.append(module_item)

        if self.group_by_preference == db.GroupByPreference.GROUP_BY_MODULES:
            exclude_assignments_module = list(filter(lambda x: x.name.lower() not in default_assignments_module_names, course.modules))

            self.files_column.load_items(exclude_assignments_module)
        else:
            self.files_column.load_items(module_items)

        self.assignment_files_column.load_items(assignment_items + course.assignments)

    def clear(self):
        self.group_by_preference = None
        self.files_column.clear()
        self.assignment_files_column.clear()

    @Slot(QPoint)
    def files_column_context_menu(self, pos: QPoint):
        if self.group_by_preference is None:
            return

        menu = QMenu(self.files_column)

        group_by_menu = menu.addMenu("Group by")

        select_group_preference_modules = create_qaction(
            name="Modules",
            checkable=True,
            checked=self.group_by_preference == db.GroupByPreference.GROUP_BY_MODULES,
            triggered=lambda: self.group_by_preference_changed.emit(db.GroupByPreference.GROUP_BY_MODULES)
        )

        select_group_preference_pages = create_qaction(
            name="Pages",
            checkable=True,
            checked=self.group_by_preference == db.GroupByPreference.GROUP_BY_PAGES,
            triggered=lambda: self.group_by_preference_changed.emit(db.GroupByPreference.GROUP_BY_PAGES)
        )

        action_group = QActionGroup(menu)
        action_group.addAction(select_group_preference_modules)
        action_group.addAction(select_group_preference_pages)

        group_by_menu.addAction(select_group_preference_pages)
        group_by_menu.addAction(select_group_preference_modules)

        menu.exec(self.files_column.mapToGlobal(pos))





from qcanvas.QtVersionHelper.QtCore import QSettings, QUrl


class _AppSettings:
    def __init__(self):
        self.settings = QSettings("QCanvas", "client")
        self.auxiliary = QSettings("QCanvas", "ui")
        self._canvas_url = self.settings.value("canvas_url", None)
        self._api_key = self.settings.value("api_key", defaultValue=None)

    @property
    def canvas_url(self) -> str | None:
        return self._canvas_url

    @canvas_url.setter
    def canvas_url(self, value: str):
        self._canvas_url = value
        self.settings.setValue("canvas_url", value)

    @property
    def canvas_api_key(self) -> str | None:
        return self._api_key

    @canvas_api_key.setter
    def canvas_api_key(self, value: str):
        self._api_key = value
        self.settings.setValue("api_key", value)

    @property
    def is_set(self):
        return self.canvas_url is not None and QUrl(self.canvas_url).isValid() and self.canvas_url is not None





import logging
from asyncio import Lock, Event


class AuthenticationException(Exception):
    pass

# httpx does have an authentication flow mechanism that allows you to also make other requests but I don't know if it
# will behave the same way as this does. I also finished this before I found out that existed.

class SelfAuthenticating:
    _logger = logging.getLogger(__name__)

    def __init__(self):
        self.authentication_lock = Lock()
        self.authentication_in_progress: Event | None = None

    async def reauthenticate(self):
        """
        Re-authenticates the client
        """

        # Acquire the lock as to prevent multiple threads from modifying self.authentication_in_progress.
        # Not that will ever happen using asyncio alone because python async is not actually multi-processed
        await self.authentication_lock.acquire()

        # If there's no authentication already in progress
        if self.authentication_in_progress is None:
            # Update authentication_in_progress and release the lock
            self.authentication_in_progress = Event()
            self.authentication_lock.release()

            # Start the authentication
            self._logger.info(f"Authenticating {self.__class__.__name__}")
            await self._authenticate()
            self._logger.info(f"Finished authenticating {self.__class__.__name__}")
            # Update the event to unblock other coroutines
            self.authentication_in_progress.set()

            # Delete authentication_in_progress to indicate that authentication is no longer in progress
            # This lock is probably not needed (due to not actually being multi-processed) but shouldn't hurt
            await self.authentication_lock.acquire()
            self.authentication_in_progress = None
            self.authentication_lock.release()
        # Authentication is already in progress
        else:
            # Get a reference to the event
            event = self.authentication_in_progress
            # Release the lock as we are not going to access it anymore
            self.authentication_lock.release()

            # Wait for the reauthentication to finish
            self._logger.info(f"Waiting for {self.__class__.__name__}")
            await event.wait()
            self._logger.info(f"Finished waiting for {self.__class__.__name__}")

    async def _authenticate(self) -> None:
        """
        Authenticates the client
        """
        raise NotImplemented
